/**
 * @file main.c
 * @brief Este projeto demonstra como converter os codigos binarios em grandezas
 * fisicas correspondentes aos sinais amostrados usando o sensor de temperatura LM61 
 * conectado no pino PORTB1 e o sensor de temperatura AN3031 integrado no Kinetis. Os parametros de 
 * conversao so os da Tabela 28-75 do Reference Manual
 * @author Wu Shin-Ting
 * @date 15/11/2020
 */

#include "derivative.h" /* include peripheral declarations */

//==================================
//Clocks
//==================================
/**
 * @brief Setar bus clock em 20,97152MHz/1
 */
void SIM_initBusClock (uint8_t OUTDIV4) {	
	SIM_CLKDIV1 &= ~SIM_CLKDIV1_OUTDIV4(0b111);  //resetar todos os bits
	SIM_CLKDIV1 |= SIM_CLKDIV1_OUTDIV4(OUTDIV4); //mascarar 1s
}

//==================================
//PORTB
//==================================
/**
 * @brief Inicializar o pino de termometro
 */
void PORTB_initTermometer(void) {
	
	// Modulo SIM: habilita o clock do modulo PORTE
	SIM_SCGC5 |= SIM_SCGC5_PORTB_MASK; 
	
	// Modulo PORT: configure o pino PTB1 para ADC0
	PORTB_PCR1 &= ~PORT_PCR_MUX (0b111); 
	
	PORTB_PCR1 |= PORT_PCR_ISF_MASK;  
}

//==================================
// ADC
//==================================
/**
 * @ configuration parameters of ADC
 */
typedef struct _ADCConfiguration {
	uint8_t sc1_diff;       // conversao unipolar (0) ou diferencial (1)
	uint8_t cfg1_adlpc;		// configuracao de baixa potencia
	uint8_t cfg1_adiv;		// divisor de frequencia do ADICLK
	uint8_t cfg1_adlsmp; 	// configuracao do tempo de amostragem
	uint8_t cfg1_mode;		// resolucao e formato da saida
	uint8_t cfg1_adiclk;	// fonte de clock de entrada
	uint8_t cfg2_muxsel; 	// selecao do canal: A (0) e B (1)
	uint8_t cfg2_adacken;	// habilicao da saida de sinal de clock assincrono
	uint8_t cfg2_adhsc;		// habilitacao de operacao em alta velocidade
	uint8_t cfg2_adlsts; 	// selecao de tempo de amostragem longo
	uint8_t sc2_adtrg;		// selecao de trigger: software (0) e hardware (1)
	uint8_t sc2_acfe; 		// habilitacao de funcao de comparacao em igualdade
	uint8_t sc2_acfgt;		// habilitacao de funcao de comparacao em desigualdade
	uint8_t sc2_acren;		// habilitacao de comparacao de uma faixa
	uint8_t sc2_dmaen; 		// habilitacao de DMA
	uint8_t sc2_refsel;		// selecao de tensoes de referencia
	uint8_t sc3_adco;		// habilitacao de amostragem continua (1) ou unica (0)
	uint8_t sc3_avge; 		// habilitacao de media de amostras por hardware
	uint8_t sc3_avgs; 		// selecao da quantidade de amostras por media
} ADCConfig;

/**
 * @brief Calibrar ADC usando a funcao de calibracao
 * 		integrada ao chip segundo o procedimento recomendado pelo fabricante.
 * @note Secao 28.4.6 de KL-25 Sub-Family Reference Manual
 * \param[in] config configuration parameters
 */
void ADC0_calibrate (ADCConfig *config)
{
	uint16_t tmp=0;
	
	// Configurar qtde de entradas analogicas (unipolar/biploar) e desabilitar o canal
	ADC0_SC1A &= ~(ADC_SC1_DIFF_MASK );

	ADC0_SC1A |= (config->sc1_diff << ADC_SC1_DIFF_SHIFT);
	
	ADC0_SC1A |= ADC_SC1_ADCH(31);
	// Configurar frequencia ADCK (<= 4MHz) e resolucao, formato de saida
	ADC0_CFG1 &= ~(ADC_CFG1_ADICLK_MASK |
			ADC_CFG1_MODE_MASK |
			ADC_CFG1_ADLSMP_MASK |
			ADC_CFG1_ADIV_MASK |
			ADC_CFG1_ADLPC_MASK);
			
	ADC0_CFG1 |= (config->cfg1_adiclk << ADC_CFG1_ADICLK_SHIFT |	
			config->cfg1_mode << ADC_CFG1_MODE_SHIFT |
			config->cfg1_adlsmp << ADC_CFG1_ADLSMP_SHIFT |
			0b11 << ADC_CFG1_ADIV_SHIFT |
			config->cfg1_adlpc << ADC_CFG1_ADLPC_SHIFT);

	// Configurar o tempo de conversao e a velocidade de operacao
	ADC0_CFG2 &= ~(ADC_CFG2_MUXSEL_MASK |
			ADC_CFG2_ADACKEN_MASK |
			ADC_CFG2_ADHSC_MASK |
			ADC_CFG2_ADLSTS_MASK);
	
	ADC0_CFG2 |= (config->cfg2_muxsel << ADC_CFG2_MUXSEL_SHIFT |
			config->cfg2_adacken << ADC_CFG2_ADACKEN_SHIFT |
			config->cfg2_adhsc << ADC_CFG2_ADHSC_SHIFT |
			config->cfg2_adlsts << ADC_CFG2_ADLSTS_SHIFT);

	// Configurar fonte de referencia, trigger, funcao de comparacao habilitada e acesso DMA
	ADC0_SC2 &= ~( ADC_SC2_ADTRG_MASK |
			ADC_SC2_ACFE_MASK  |
			ADC_SC2_ACFGT_MASK |
			ADC_SC2_ACREN_MASK | 
			ADC_SC2_DMAEN_MASK |
			ADC_SC2_REFSEL_MASK);
	
	ADC0_SC2 |= (config->sc2_acfe << ADC_SC2_ACFE_SHIFT  |
			0b1 << ADC_SC2_ACFGT_SHIFT |
			0b1 << ADC_SC2_ACREN_SHIFT | 
			config->sc2_dmaen << ADC_SC2_DMAEN_SHIFT |
			0b00 << ADC_SC2_REFSEL_SHIFT);
	
	// Setar valores de comparacao
	ADC0_CV1 = 0x1234u;
	ADC0_CV2 = 0x5678u;

	// Configurar modo de amostrage e media das amostras (32)
	ADC0_SC3 &= ~( ADC_SC3_ADCO_MASK |
			ADC_SC3_AVGE_MASK |
			ADC_SC3_AVGS_MASK);
	
	ADC0_SC3 |= (0b0 << ADC_SC3_ADCO_SHIFT |
			0b1 << ADC_SC3_AVGE_SHIFT |
			0b11 << ADC_SC3_AVGS_SHIFT);
	
	// Calibrar
	ADC0_SC3 |= (ADC_SC3_CAL_MASK);			// inicializa a calibracao	
	while (ADC0_SC3 & ADC_SC3_CAL_MASK)  {}	///< aguarda a calibracao

	// Gere as correcos de erros de ganho (positivo)
	tmp = ADC0_CLP0 + ADC0_CLP1 + ADC0_CLP2 + 
			ADC0_CLP3 + ADC0_CLP4 + ADC0_CLPS;	// soma PS
	tmp = tmp/2;								// divide por 2
	tmp |= (uint16_t)(0x8000);					// setar o bit mais significativo
	ADC0_PG = tmp;								// plus-side gain
	
	// Somente util para modos diferenciais 
	tmp = ADC0_CLM0 + ADC0_CLM1 + ADC0_CLM2 + 
			ADC0_CLM3 + ADC0_CLM4 + ADC0_CLMS;	// soma PS
	tmp = tmp/2;								// divide por 2
	tmp |= (uint16_t)(0x8000);					// setar o bit mais significativo
	ADC0_MG = tmp;								// minus side gain
	
	// A compensacao do erro de offset eh atribuida em ADC0_OFS	
}

/*!
 * @brief Configura o modo de opera&ccedil;&atilde;o do m&oacute;dulo ADC. Para compactar
 * 			os diversos par&acirc;metros de configura&ccedil;&atilde;o numa vari&aacute;vel
 * 			foi definida um novo tipo de dado (struct) ADCConf em adc.h.
 * \note Um exemplo de configura&ccedil;&atilde;o &eacute; dado em 
 * 		 ftp://ftp.dca.fee.unicamp.br/pub/docs/ea871/ARM/KLQRUG.pdf#page=117
 * \param[in] config parametros de configuracao.
 */
void ADC0_configure(ADCConfig *config) {
	
	// Configurar qtde de entradas analogicas (unipolar/biploar)
	ADC0_SC1A &= ~(ADC_SC1_DIFF_MASK );

	ADC0_SC1A |= (config->sc1_diff << ADC_SC1_DIFF_SHIFT);
	
	ADC0_SC1A |= ADC_SC1_ADCH(31);
	// Configurar frequencia ADCK e resolucao, formato de saida
	ADC0_CFG1 &= ~(ADC_CFG1_ADICLK_MASK |
			ADC_CFG1_MODE_MASK |
			ADC_CFG1_ADLSMP_MASK |
			ADC_CFG1_ADIV_MASK |
			ADC_CFG1_ADLPC_MASK);
			
	ADC0_CFG1 |= (config->cfg1_adiclk << ADC_CFG1_ADICLK_SHIFT |	
			config->cfg1_mode << ADC_CFG1_MODE_SHIFT |
			config->cfg1_adlsmp << 	ADC_CFG1_ADLSMP_SHIFT |
			config->cfg1_adiv << ADC_CFG1_ADIV_SHIFT |
			config->cfg1_adlpc << ADC_CFG1_ADLPC_SHIFT);

	// Configurar o tempo de conversao e a velocidade de operacao
	ADC0_CFG2 &= ~(ADC_CFG2_MUXSEL_MASK |
			ADC_CFG2_ADACKEN_MASK |
			ADC_CFG2_ADHSC_MASK |
			ADC_CFG2_ADLSTS_MASK);
	
	ADC0_CFG2 |= (config->cfg2_muxsel << ADC_CFG2_MUXSEL_SHIFT |
			config->cfg2_adacken << ADC_CFG2_ADACKEN_SHIFT |
			config->cfg2_adhsc << ADC_CFG2_ADHSC_SHIFT |
			config->cfg2_adlsts << ADC_CFG2_ADLSTS_SHIFT);

	// Configurar fonte de referencia, trigger, funcao de comparacao e acesso DMA
	ADC0_SC2 &= ~( ADC_SC2_ADTRG_MASK |
			ADC_SC2_ACFE_MASK  |
			ADC_SC2_ACFGT_MASK |
			ADC_SC2_ACREN_MASK | 
			ADC_SC2_DMAEN_MASK |
			ADC_SC2_REFSEL_MASK);
	
	ADC0_SC2 |= ( config->sc2_adtrg << ADC_SC2_ADTRG_SHIFT |
			config->sc2_acfe << ADC_SC2_ACFE_SHIFT  |
			config->sc2_acfgt << ADC_SC2_ACFGT_SHIFT |
			config->sc2_acren << ADC_SC2_ACREN_SHIFT | 
			config->sc2_dmaen << ADC_SC2_DMAEN_SHIFT |
			config->sc2_refsel << ADC_SC2_REFSEL_SHIFT);
	
	// Configurar modo de amostrage e media das amostras 
	ADC0_SC3 &= ~( ADC_SC3_ADCO_MASK |
			ADC_SC3_AVGE_MASK |
			ADC_SC3_AVGS_MASK);
	
	ADC0_SC3 |= (config->sc3_adco << ADC_SC3_ADCO_SHIFT |
			config->sc3_avge << ADC_SC3_AVGE_SHIFT |
			config->sc3_avgs << ADC_SC3_AVGS_SHIFT);
}

/**
 * @brief Inicializar ADC
 */
void ADC0_init(ADCConfig *config) {
	//Habilitar o clock
	SIM_SCGC6 |= SIM_SCGC6_ADC0_MASK;  
	
	//Calibrar ADC0
	ADC0_calibrate(config);
	
	//Configurar ADC0
	ADC0_configure(config);
}

/**
 * @brief Setar trigger por hardware
 */
void ADC0_initADHWT (uint8_t trigger) {
	//Configurar o trigger por hardware
	SIM_SOPT7 &= SIM_SOPT7_ADC0TRGSEL(0b111);      // resetar o campo
	
	SIM_SOPT7 |= (SIM_SOPT7_ADC0ALTTRGEN_MASK      // trigger alternativo de TPM1_CH0 e TPM1_CH1
			| SIM_SOPT7_ADC0TRGSEL(trigger)) ;      
	
	SIM_SOPT7 &= ~SIM_SOPT7_ADC0PRETRGSEL_MASK;	   // pre-trigger A

	// Habilitar trigger por hardware
	ADC0_SC2 |= ADC_SC2_ADTRG_MASK;
}

/**
 * @brief Setar trigger por software
 */
void ADC0_initSC1A (void) {
	// Habilitar trigger por software
	ADC0_SC2 &= ~ADC_SC2_ADTRG_MASK;
}

/*!
 * @brief Habilita interrupcao de conversao completa do ADC
 */
void ADC0_enableInterrup (void) {
	ADC0_SC1A |= ADC_SC1_AIEN_MASK;           
}

/*!
 * @brief Desabilita a interrupcao de conversao completa do ADC
 */
void ADC0_disableInterrup(void) {
	ADC0_SC1A &= ~ADC_SC1_AIEN_MASK;           
}

/**
 * @brief Selecionar o canal de entrada analogica unipolar
 */
void ADC0_selSingleChannel (uint8_t canal) {
	ADC0_SC1A &= ~ADC_SC1_ADCH (0b11111);
	
	ADC0_SC1A |= ADC_SC1_ADCH (ADC_SC1_ADCH(canal));
}

//====================================
// NVIC
//====================================
void NVIC_enableADC0IRQ (char priority) {
	NVIC_ISER |= 1 << (31-16);         	// NVIC_ISER[22]=1 (habilita IRQ22)
	NVIC_ICPR |= 1 << (31-16);         	// NVIC_ICPR[22]=1 (limpa as pendncias)
	NVIC_IPR3 |= NVIC_IP_PRI_15(priority << 6); // 22/4 = 5
}

//=====================================
// Application
//=====================================
#define MAX 0x3ff			// resolucao 10 bits 
#define VREFH 3300			// tensao de refeencia alta (mV)
#define VREFL 0.0 			// tensao de referencia baixa(mV)


ADCConfig config = {
		.sc1_diff=0,      			// conversao unipolar (0) ou diferencial (1)
		.cfg1_adlpc=0,				// potencia normal
		.cfg1_adiv=0b00,			// 1
		.cfg1_adlsmp=0, 			// tempo de amostragem curto
		.cfg1_mode=0b00,			// 8 bits (unipolar)_/9 bits (bipolar)
		.cfg1_adiclk=0b00,			// bus clock
		.cfg2_muxsel=0, 			// canal A (0)
		.cfg2_adacken=0,			// desabilitar a saida de sinal de clock assincrono
		.cfg2_adhsc=0,				// desabilitar operacao em alta velocidade
		.cfg2_adlsts=0b00, 			// tempo de amostragem mais longo
		.sc2_adtrg=0,				// trigger por software
		.sc2_acfe=0, 		        // desabilitar funcao de comparacao em igualdade
		.sc2_acfgt=0,				// desabilitar funcao de comparacao em desigualdade
		.sc2_acren=0,				// desabilitar comparacao de uma faixa
		.sc2_dmaen=0, 				// habilitar de DMA
		.sc2_refsel=0b00,			// tensoes de referencia VREFH e VREFL
		.sc3_adco=0,				// amostragem unica
		.sc3_avge=0, 				// desabilitar media de amostras por hardware
		.sc3_avgs=0b00 				// 4 amostras
};

int main(void)
{
	uint32_t valor, temperaturaLM61, temperatura3031;
	uint16_t result0A;
	float m, valorF;
	
	SIM_initBusClock (0b000);           // OUTDIV4 = 1 
	
//	config.cfg1_adiv=0b00;			// 1
//	config.cfg1_mode=0b10;			// 10 bits 
//	config.sc3_adco=0;				// amostragem unica
	
//	config.cfg1_adiv=0b10;			// 4
//	config.cfg1_mode=0b01;			// 12 bits 
//	config.sc2_adtrg=1;				// trigger por hardware
//	config.sc3_adco=0;				// amostragem unica

	config.cfg1_adiv=0b10;			// 4
	config.cfg1_adlsmp=0; 			// longo
	config.cfg1_mode=0b01;			// 12 bits 
	config.cfg2_adhsc=1;			// habilitar operacao em alta velocidade
	config.cfg2_adlsts=0b00; 		// 24 ADCK ciclos
	config.sc2_adtrg=1;				// trigger por hardware
	config.sc3_avge=1; 				// habilitar media de amostras por hardware
	config.sc3_avgs=0b10;           // 16 amostras  
	config.sc3_adco=0;				// amostragem unica
	

	ADC0_init(&config);				// inicializar ADC com os parametros especificados
	
	PORTB_initTermometer();			// inicializar o pino onde o sensor esta conectado
	
	ADC0_selSingleChannel (0b01001); // canal de sensor de temperatura (AD9)
	
	for (;;) {
			GPIOE_PTOR = (1<<23);
			while (!(ADC0_SC1A & ADC_SC1_COCO_MASK)) {} //polling
			result0A = ADC0_RA;
 			valor = (result0A * (VREFH-VREFL))/MAX;  // em mV
			temperaturaLM61 = (valor - 600);         // em decigraus em Celsius

			ADC0_selSingleChannel (0b11010); // canal de sensor de temperatura (AD26)
			while (!(ADC0_SC1A & ADC_SC1_COCO_MASK)) {} //polling
			result0A = ADC0_RA;
			valorF = ((float)(result0A) * (VREFH-VREFL))/MAX;                  

			if (valorF < 703.125) m = 1.646;               // inclinacao do segmento
			else m = 1.769;

			temperatura3031 = (25. - ((valorF - 703.125)/m)*0.001)*10; // em decigraus Celsius

			ADC0_selSingleChannel (0b01001); // canal de sensor de temperatura (AD9)
					
	}
	
	return 0;
}
